function [EXP]=Voronoi_Loop(EXP)

[V,C]=VoronoiBounded(EXP.Pose(1,:)',EXP.Pose(2,:)', EXP.Workspace); %Compute Voronoi's Cells and Vertexes

C=AdjustNumericalBugs(V,C); % correct numerical errors in vertices

%-- Compute vertexes of Voronoi cells
EXP.Voronoi.All_vertices=V;
for j=1:EXP.Robots
    EXP.Voronoi.Cell(j).Vertex=V(C{j},1:2);
end
%------

%-- Compute Adjacency matrix
EXP = Find_Neighbors(C,EXP);
%------

%-- Compute Cell Areas 
for i=1:EXP.Robots
    [~,~,area] = PolyCentroid(V(C{i},1),V(C{i},2));
    EXP.Voronoi.Cell(i).Area=abs(area);
end
%------

%-- Compute centroids of Voronoi cells
for j=1:EXP.Robots
    if isfield(EXP.Voronoi.Cell(j),'Enable_centroid_computation') && (EXP.Voronoi.Cell(j).Enable_centroid_computation)
        [centroid_x,centroid_y] = PolyCentroid(V(C{j},1),V(C{j},2)); %Compute Voronoi's Centroids
        EXP.Voronoi.Cell(j).Centroid=[centroid_x, centroid_y];
    end
end
%------
end

function EXP = Find_Neighbors(C,EXP) 
% Computes adjacency sparse matrix from Voronoi vertices C.
% M(i,j) = 1 = M(j,i) if i,j are are neighbors, 0 otherwise.
% Set neighbors and corresponding extremal points of the Voronoi cell for each agent

N=length(C); % number of agents
M=sparse(N,N);
for i=1:N
    EXP.Voronoi.Cell(i).Neighbors={};
    EXP.Voronoi.Cell(i).Neighbors_vertices={};
end
for i=1:N
    Ci=C{i}(1:end-1); % remove last entry because duplicated
    for j=i+1:N
        Extreme_vertices=[0 0;0 0]; % coordinates of the two etremal vertices of the segment
        Cj=C{j}(1:end-1); % remove last entry because duplicated
        s=0;
        for i1=1:length(Ci)  % check intersection between Ci and Cj
            for i2=1:length(Cj)
                if Ci(i1)==Cj(i2)
                    s=s+1;
                    Extreme_vertices(s,:)=EXP.Voronoi.All_vertices(Ci(i1),:);
                    continue;
                end
            end
        end
        if(s>1)
            EXP.Voronoi.Cell(i).Neighbors(end+1)={j}; % set neighbor index
            EXP.Voronoi.Cell(j).Neighbors(end+1)={i}; % set neighbor index
            EXP.Voronoi.Cell(i).Neighbors_vertices(end+1)={Extreme_vertices}; % set neighbor vertices
            EXP.Voronoi.Cell(j).Neighbors_vertices(end+1)={Extreme_vertices}; % set neighbor vertices
            M(i,j)=1;
            M(j,i)=1;
        end
    end
end
EXP.Voronoi.AdjM=M;
end

function [Cx,Cy,A] = PolyCentroid(X,Y)
% POLYCENTROID returns the coordinates for the centroid of polygon with vertices X,Y
% The centroid of a non-self-intersecting closed polygon defined by n vertices (x0,y0), (x1,y1), ..., (xn?1,yn?1) is the point (Cx, Cy), where
% In these formulas, the vertices are assumed to be numbered in order of their occurrence along the polygon's perimeter, and the vertex ( xn, yn ) is assumed to be the same as ( x0, y0 ). Note that if the points are numbered in clockwise order the area A, computed as above, will have a negative sign; but the centroid coordinates will be correct even in this case.http://en.wikipedia.org/wiki/Centroid
% A = polyarea(X,Y)

Xa = [X(2:end);X(1)];
Ya = [Y(2:end);Y(1)];

A = 1/2*sum(X.*Ya-Xa.*Y); %signed area of the polygon, positive if vertices in COUNTERCLOCKWISE order

Cx = (1/(6*A)*sum((X + Xa).*(X.*Ya-Xa.*Y)));
Cy = (1/(6*A)*sum((Y + Ya).*(X.*Ya-Xa.*Y)));
end

function [V,C]=VoronoiBounded(x,y, crs)
% VORONOIBOUNDED computes the Voronoi cells about the points (x,y) inside
% the bounding box (a polygon) crs.  If crs is not supplied, an
% axis-aligned box containing (x,y) is used.

    bnd=[min(x) max(x) min(y) max(y)]; %data bounds
    if nargin < 3
        crs=double([bnd(1) bnd(4);bnd(2) bnd(4);bnd(2) bnd(3);bnd(1) bnd(3);bnd(1) bnd(4)]);
    end

    rgx = max(crs(:,1))-min(crs(:,1));
    rgy = max(crs(:,2))-min(crs(:,2));
    rg = max(rgx,rgy);
    midx = (max(crs(:,1))+min(crs(:,1)))/2;
    midy = (max(crs(:,2))+min(crs(:,2)))/2;

    % add 4 additional edges
    xA = [x; midx + [0;0;-5*rg;+5*rg]];
    yA = [y; midy + [-5*rg;+5*rg;0;0]];
    

    [vi,ci]=voronoin([xA,yA],{'Qbb','Qz'});

    % remove the last 4 cells
    C = ci(1:end-4);
    V = vi;
    % use Polybool to crop the cells
    %Polybool for restriction of polygons to domain.

    for ij=1:length(C)
        % thanks to http://www.mathworks.com/matlabcentral/fileexchange/34428-voronoilimit
        % first convert the contour coordinate to clockwise order:
        [X2, Y2] = poly2cw(V(C{ij},1),V(C{ij},2));
        [xb, yb] = polybool('intersection',crs(:,1),crs(:,2),X2,Y2);
        ix=nan(1,length(xb));
        for il=1:length(xb)
            if any(V(:,1)==xb(il)) && any(V(:,2)==yb(il))
                ix1=find(V(:,1)==xb(il));
                ix2=find(V(:,2)==yb(il));
                for ib=1:length(ix1)
                    if any(ix1(ib)==ix2)
                        ix(il)=ix1(ib);
                    end
                end
                if isnan(ix(il))==1
                    lv=length(V);
                    V(lv+1,1)=xb(il);
                    V(lv+1,2)=yb(il);
                    ix(il)=lv+1;
                end
            else
                lv=length(V);
                V(lv+1,1)=xb(il);
                V(lv+1,2)=yb(il);
                ix(il)=lv+1;
            end
        end
        C{ij}=ix;
   
    end
end

% remove bugs due to numerical approximations in vertices V
% Leave V unchanged but changes C
function Cnew=AdjustNumericalBugs(V,C) 
change=[];
toll=1e-16; % tolerance
N_vertices=size(V,1);
for i=1:N_vertices
    if isnan(V(i,:)), continue; end
    for j=i+1:N_vertices
        if ((V(i,1)-V(j,1))^2+(V(i,2)-V(j,2))^2)<toll  % faster than using norm()^2
            V(j,:)=[nan;nan];  % mark the duplicate vertex as Nan
            change=[change ; i j];
        end
    end
end
if isempty(change), Cnew=C; return; end
change_size=size(change,1);
for i=1:size(C,1)
    c=C{i};
    for j=1:length(c)
        cj=c(j);
        for k=1:change_size
            if change(k,2)==cj, c(j)=change(k,1); continue; end
        end
    end
    Cnew{i}=c;
end
end
